'''
The code in this file has been adapted from:
https://github.com/lingjzhu/charsiu
Authors: lingjzhu, pkadambi
MIT license
'''
import sys
import torch
from itertools import groupby

sys.path.append('src/')
import numpy as np
# sys.path.insert(0,'src')
from Wav2TextGrid.aligner_core.alignermodel import Wav2Vec2ForFrameClassificationSAT
from Wav2TextGrid.aligner_core.utils import seq2duration, forced_align, duration2textgrid, word2textgrid
from Wav2TextGrid.aligner_core.processors import CharsiuPreprocessor_en



class charsiu_aligner:

    def __init__(self,
                 lang='en',
                 sampling_rate=16000,
                 device=None,
                 recognizer=None,
                 processor=None,
                 resolution=0.01):

        self.lang = lang

        if processor is not None:
            self.processor = processor
        else:
            self.charsiu_processor = CharsiuPreprocessor_en()

        self.resolution = resolution

        self.sr = sampling_rate

        self.recognizer = recognizer

        if device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            self.device = device

    def _freeze_model(self):
        self.aligner.eval().to(self.device)
        if self.recognizer is not None:
            self.recognizer.eval().to(self.device)

    def align(self, audio, text):
        raise NotImplementedError()

    def serve(self, audio, save_to, output_format='variable', text=None):
        raise NotImplementedError()

    def _to_textgrid(self, phones, save_to):
        '''
        Convert output tuples to a textgrid file

        Parameters
        ----------
        phones : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        '''
        duration2textgrid(phones, save_path=save_to)
        print('Alignment output has been saved to %s' % (save_to))

    def _to_tsv(self, phones, save_to):
        '''
        Convert output tuples to a tab-separated file

        Parameters
        ----------
        phones : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        '''
        with open(save_to, 'w') as f:
            for start, end, phone in phones:
                f.write('%s\t%s\t%s\n' % (start, end, phone))
        print('Alignment output has been saved to %s' % (save_to))


class xVecSAT_forced_aligner(charsiu_aligner):
    def __init__(self, aligner, satvector_size, sil_threshold=4, **kwargs):
        super(xVecSAT_forced_aligner, self).__init__(**kwargs)
        self.aligner = Wav2Vec2ForFrameClassificationSAT.from_pretrained(aligner, satvector_size=satvector_size)
        self.sil_threshold = sil_threshold
        self._freeze_model()

    def align(self, audio, text, ixvector, target_phones=None, return_logits=False, TEMPERATURE=1):
        '''
        Perform forced alignment

        Parameters
        ----------
        audio : np.ndarray [shape=(n,)]
            time series of speech signal
        text : str
            The transcription

        Returns
        -------
        A tuple of aligned phones in the form (start_time, end_time, phone)

        '''
        audio = self.charsiu_processor.audio_preprocess(audio, sr=self.sr)
        audio = torch.Tensor(audio).unsqueeze(0).to(self.device)
        phones, words = self.charsiu_processor.get_phones_and_words(text)

        if target_phones is not None:
            try:
                if not (np.array(phones)==np.array(target_phones)).all():
                    print('target')
                    print(target_phones)
                    print('g2p')
                    print(phones)
            except:
                print('text')
                print(text)
                print('target')
                print(target_phones)
                print('g2p')
                phnsg2p = [item for t in phones for item in t]
                print(phnsg2p)
            phones=target_phones

        phone_ids = self.charsiu_processor.get_phone_ids(phones)

        with torch.no_grad():
            out = self.aligner(input_values=audio, ixvector=ixvector)
        cost = torch.softmax(out.logits, dim=-1).detach().cpu().numpy().squeeze()

        sil_mask = self._get_sil_mask(cost)

        nonsil_idx = np.argwhere(sil_mask != self.charsiu_processor.sil_idx).squeeze()
        if nonsil_idx is None:
            raise Exception("No speech detected! Please check the audio file!")

        aligned_phone_ids = forced_align(cost[nonsil_idx, :], phone_ids[1:-1])

        aligned_phones = [self.charsiu_processor.mapping_id2phone(phone_ids[1:-1][i]) for i in aligned_phone_ids]

        pred_phones = self._merge_silence(aligned_phones, sil_mask)

        pred_phones = seq2duration(pred_phones, resolution=self.resolution)

        pred_words = self.charsiu_processor.align_words(pred_phones, phones, words)
        return pred_phones, pred_words

    def serve(self, audio, text, ixvector, save_to, target_phones=None, output_format='textgrid', verbose=False):
        '''
         A wrapper function for quick inference

        Parameters
        ----------
        audio : TYPE
            DESCRIPTION.
        text : TYPE, optional
            DESCRIPTION. The default is None.
        output_format : str, optional
            Output phone-taudio alignment as a "tsv" or "textgrid" file.
            The default is 'textgrid'.

        Returns
        -------
        None.

        '''
        phones, words = self.align(audio, text, ixvector, target_phones=target_phones)

        if output_format == 'tsv':
            if save_to.endswith('.tsv'):
                save_to_phone = save_to.replace('.tsv', '_phone.tsv')
                save_to_word = save_to.replace('.tsv', '_word.tsv')
            else:
                save_to_phone = save_to + '_phone.tsv'
                save_to_word = save_to + '_word.tsv'

            self._to_tsv(phones, save_to_phone)
            self._to_tsv(words, save_to_word)

        elif output_format == 'textgrid':
            self._to_textgrid(phones, words, save_to)
        else:
            raise Exception('Please specify the correct output format (tsv or textgird)!')

    def _to_textgrid(self, phones, words, save_to):
        '''
        Convert output tuples to a textgrid file

        Parameters
        ----------
        phones : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        '''
        word2textgrid(phones, words, save_path=save_to)
        # print('Alignment output has been saved to %s'%(save_to))

    def _get_sil_mask(self, cost):
        # single out silent intervals

        preds = np.argmax(cost, axis=-1)
        sil_mask = []
        for key, group in groupby(preds):
            group = list(group)
            if (key == self.charsiu_processor.sil_idx and len(group) < self.sil_threshold):
                sil_mask += [-1 for i in range(len(group))]
            else:
                sil_mask += group

        return np.array(sil_mask)

    def _merge_silence(self, aligned_phones, sil_mask):
        # merge silent and non-silent intervals
        pred_phones = []
        count = 0
        for i in sil_mask:
            if i == self.charsiu_processor.sil_idx:
                pred_phones.append('[SIL]')
            else:
                pred_phones.append(aligned_phones[count])
                count += 1
        assert len(pred_phones) == len(sil_mask)
        return pred_phones

if __name__=='__main__':
    aligner = xVecSAT_forced_aligner('./wav2textgrid_model', satvector_size=150)



import torch
import torch.nn as nn
from transformers import Wav2Vec2ForCTC, Wav2Vec2ForPreTraining, Wav2Vec2Model
from transformers.modeling_outputs import CausalLMOutput, MaskedLMOutput


class Wav2Vec2ForFrameClassificationSAT(Wav2Vec2ForCTC):
    def __init__(self, config, satvector_size=64):
        super().__init__(config)

        self.wav2vec2 = Wav2Vec2Model(config)
        self.dropout = nn.Dropout(config.final_dropout)

        if config.vocab_size is None:
            raise ValueError(
                f"You are trying to instantiate {self.__class__} with a configuration that "
                "does not define the vocabulary size of the language model head. Please "
                "instantiate the model as follows: `Wav2Vec2ForCTC.from_pretrained(..., vocab_size=vocab_size)`. "
                "or define `vocab_size` of your model's configuration."
            )
        output_hidden_size = (
            config.output_hidden_size if hasattr(config, "add_adapter") and config.add_adapter else config.hidden_size
        )
        self.lm_head = nn.Linear(output_hidden_size, config.vocab_size)

        self.lm_head_ivec = nn.Linear(output_hidden_size + satvector_size, config.vocab_size )

        #transfer the weights
        orig_weights = self.lm_head.weight.data.detach().numpy()
        orig_bias = self.lm_head.bias.data.detach().numpy()

        weight_shape = self.lm_head.weight.shape
        bias_shape = self.lm_head.bias.shape

        weights = self.lm_head_ivec.weight.data.detach().numpy()
        biases = self.lm_head_ivec.bias.data.detach().numpy()

        weights[:weight_shape[0], :weight_shape[1]] = orig_weights
        biases[:bias_shape[0]] = orig_bias

        self.lm_head_ivec.weight.data = torch.tensor(weights)
        self.lm_head_ivec.bias.data = torch.tensor(biases)


        # Initialize weights and apply final processing
        self.post_init()

    def forward(
            self,
            input_values,
            attention_mask=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
            labels=None,
            ixvector=None,
    ):

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.wav2vec2(
            input_values,
            attention_mask=attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = outputs[0]


        hidden_states = self.dropout(hidden_states)

        #framewise append i/xvector
        # satvector = ixvector.unsqueeze(dim=1).repeat(1, hidden_states.shape[1], 1)
        satvector = ixvector.unsqueeze(dim=1).repeat(1, hidden_states.shape[1], 1)
        hidden_states = torch.cat([hidden_states, satvector], dim=-1)


        logits = self.lm_head_ivec(hidden_states)
        timesteps = logits.shape[1]
        loss = None
        if labels is not None:
            labels = labels[:,:timesteps]
            if labels.max() >= self.config.vocab_size:
                raise ValueError(f"Label values must be <= vocab_size: {self.config.vocab_size}")

            # retrieve loss input_lengths from attention_mask
            attention_mask = (
                attention_mask if attention_mask is not None else torch.ones_like(input_values, dtype=torch.long)
            )
            input_lengths = self._get_feat_extract_output_lengths(attention_mask.sum(-1)).to(torch.long)

            loss = torch.nn.functional.cross_entropy(logits.view(-1, logits.size(2)), labels.flatten(),
                                                     reduction="mean")

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutput(
            loss=loss, logits=logits, hidden_states=outputs.hidden_states, attentions=outputs.attentions
        )


#!/usr/bin/env python3
import glob
import os
import pickle as pkl
import torch
from Wav2TextGrid.aligner_core.xvec_extractor import xVecExtractor
from Wav2TextGrid.aligner_core.aligner import xVecSAT_forced_aligner
import argparse

def align_file(args):
    xvx = xVecExtractor(method='xvector')

    xvector = xvx.extract_xvector(args.wavfile_or_dir)
    xvector = xvector[0][0].view(1, -1)
    if torch.cuda.is_available():
        xvector = xvector.cuda()

    aligner = xVecSAT_forced_aligner('pkadambi/Wav2TextGrid', satvector_size=512)
    transcript = open(args.transcriptfile_or_dir, 'r').readlines()[0]
    transcript = transcript.replace('\n', '')
    aligner.serve(audio=args.wavfile_or_dir, text=transcript, save_to=args.outfile_or_dir, ixvector=xvector)


def align():
    parser = argparse.ArgumentParser()
    parser.add_argument('wavfile_or_dir', type=str)
    parser.add_argument('transcriptfile_or_dir', type=str)
    parser.add_argument('outfile_or_dir', default=str)
    args = parser.parse_args()

    if os.path.isdir(args.wavfile_or_dir):
        align_dirs(args)
    else:
        align_file(args)


def align_dirs(args):
    # Get list of .wav files in directory1 and its subdirectories
    wav_files = glob.glob(os.path.join(args.wavfile_or_dir, '**/*.wav'), recursive=True)
    # Get list of .lab files in directory2 and its subdirectories
    lab_files = glob.glob(os.path.join(args.transcriptfile_or_dir, '**/*.lab'), recursive=True)
    success_count = 0
    failure_count = 0
    missing_lab_files = []
    # Iterate over .wav files
    for wav_file in wav_files:
        # Generate corresponding .lab file path
        rel_path = os.path.relpath(wav_file, args.wavfile_or_dir)
        lab_file = os.path.join(args.transcriptfile_or_dir, os.path.splitext(rel_path)[0] + '.lab')

        # Check if .lab file exists
        if os.path.exists(lab_file):
            try:
                # Align .wav and .lab files
                align(wav_file, lab_file)
                success_count += 1
            except Exception as e:
                print(f"Alignment failed for {wav_file}: {e}")
                failure_count += 1
        else:
            missing_lab_files.append(lab_file)
            print(f"Did not find transcript at {lab_file} for wav file {wav_file}")

        # Write to alignment log
    with open(os.path.join(args.output_directory_or_filepath, 'alignment.log'), 'w') as log_file:
        log_file.write(f"Successfully aligned: {success_count}\n")
        log_file.write(f"Alignment failures: {failure_count}\n")
        if missing_lab_files:
            log_file.write("\nMissing transcript files:\n")
            for missing_lab_file in missing_lab_files:
                log_file.write(f"- {missing_lab_file}\n")

    pass


if __name__=='__main__':
    align()
    # xvx = xVecExtractor(method='xvector')
    #
    # xvector = xvx.extract_xvector('./examples/test.wav')
    #
    # aligner = xVecSAT_forced_aligner('./wav2textgrid_model', satvector_size=150)
    # transcript = open('./examples/test.lab', 'r').readlines()[0]
    # transcript = transcript.replace('\n', '')
    # xvector = xvector[0][0].view(1, -1)
    #
    # if CUDA_AVAILABLE:
    #     xvector = xvector.cuda()
    #
    # # aligner.serve(audio='./examples/test.wav', text=transcript, save_to='./examples/test.TextGrid', ixvector=xvector)
    # aligner.serve(audio='./examples/test.wav', text=transcript, save_to='./examples/test.TextGrid', ixvector=xvector)